// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_bit.h
 *  \brief      MGL ビットフラグ定義
 *  \date       Since: January 8, 2021. 10:07:57 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_BIT_H_1610068077
#define INCGUARD_MGL_BIT_H_1610068077

#include <cstdint>
#include <type_traits>

namespace MGL
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ビットフラグを生成
 *  \tparam     T   生成する型
 *  \param[in]  bit ビット番号
 *  \return     引数のビット番号をオンにしたT型の値
 */
/* ------------------------------------------------------------------------- */
template <typename T>
constexpr T MakeBit(uint8_t bit) noexcept
{
    return static_cast<T>(1 << bit);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      enum classで定義した列挙型に対応したビットフラグを扱うクラス
 *  \tparam     T   enum classで定義した型
 */
/* ------------------------------------------------------------------------- */
template <class T, typename FlagType = uint32_t>
class EnumBitFlags
{
public:
    //! T型の基底の型
    using UnderlyingType = typename std::underlying_type<T>::type;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags() noexcept
        : _value(0)
    {
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     *  \param[in]  value   設定する値
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags(T value) noexcept
        : _value(FlagType(1 << UnderlyingType(value)))
    {
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     *  \param[in]  value   設定する値
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags(FlagType value) noexcept
        : _value(value)
    {
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      フラグの値を取得
     *  \return     フラグの値
     */
    /* ------------------------------------------------------------------------- */
    constexpr FlagType GetValue() const noexcept
    {
        return _value;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      指定した値を保持しているかを取得
     *  \param[in]  value   T型の値
     *  \retval     true    保持している
     *  \retval     false   保持していない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool Has(T value) const noexcept
    {
        return _value & FlagType(1 << UnderlyingType(value));
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      指定した値をセットする
     *  \param[in]  value   T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &Set(T value) noexcept
    {
        _value |= FlagType(1 << UnderlyingType(value));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      指定した値をクリアする
     *  \param[in]  value   T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &Clear(T value) noexcept
    {
        _value &= ~FlagType(1 << UnderlyingType(value));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      全てのフラグをクリアする
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &Clear() noexcept
    {
        _value = 0;
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      指定した値のフラグを反転させる
     *  \param[in]  value   T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &Flip(T value) noexcept
    {
        _value ^= FlagType(1 << UnderlyingType(value));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      全ての値を反転させる
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &Flip() noexcept
    {
        _value = ~_value;
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      bool型へのキャスト
     *  \retval     true    値が非ゼロである
     *  \retval     false   値がゼロである
     */
    /* ------------------------------------------------------------------------- */
    explicit constexpr operator bool() const noexcept
    {
        return _value != 0;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      否定演算子
     *  \retval     true    値がゼロである
     *  \retval     false   値が非ゼロである
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool operator!() const noexcept
    {
        return _value == 0;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との代入演算子
     *  \param[in]  rhs     T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator=(T rhs) noexcept
    {
        _value = FlagType(1 << UnderlyingType(rhs));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との|=演算子
     *  \param[in]  rhs     T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator|=(T rhs) noexcept
    {
        _value |= FlagType(1 << UnderlyingType(rhs));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      自身の型との|=演算子
     *  \param[in]  rhs     自身の型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator|=(const EnumBitFlags &rhs) noexcept
    {
        _value |= rhs._value;
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との&=演算子
     *  \param[in]  rhs     T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator&=(T rhs) noexcept
    {
        _value &= FlagType(1 << UnderlyingType(rhs));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      自身の型との&=演算子
     *  \param[in]  rhs     自身の型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator&=(const EnumBitFlags &rhs) noexcept
    {
        _value &= rhs._value;
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との^=演算子
     *  \param[in]  rhs     T型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator^=(T rhs) noexcept
    {
        _value ^= FlagType(1 << UnderlyingType(rhs));
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      自身の型との^=演算子
     *  \param[in]  rhs     自身の型の値
     *  \return     自身の参照
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags &operator^=(const EnumBitFlags &rhs) noexcept
    {
        _value ^= rhs._value;
        return *this;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との|演算子
     *  \param[in]  rhs     T型の値
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator|(T rhs) noexcept
    {
        return EnumBitFlags(_value | FlagType(1 << UnderlyingType(rhs)));
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      自身の型との|演算子
     *  \param[in]  rhs     自身の型の値
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator|(const EnumBitFlags &rhs) noexcept
    {
        return EnumBitFlags(_value | rhs._value);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との&演算子
     *  \param[in]  rhs     T型の値
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator&(T rhs) noexcept
    {
        return EnumBitFlags(_value & FlagType(1 << UnderlyingType(rhs)));
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      自身の型との&演算子
     *  \param[in]  rhs     自身の型の値
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator&(const EnumBitFlags &rhs) noexcept
    {
        return EnumBitFlags(_value & rhs._value);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      T型との^演算子
     *  \param[in]  rhs     T型の値
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator^(T rhs) noexcept
    {
        return EnumBitFlags(_value ^ FlagType(1 << UnderlyingType(rhs)));
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      自身の型との^演算子
     *  \param[in]  rhs     自身の型の値
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator^(const EnumBitFlags &rhs) noexcept
    {
        return EnumBitFlags(_value ^ rhs._value);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ~演算子(全てのビットの反転）
     *  \return     演算結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr EnumBitFlags operator~() noexcept
    {
        return EnumBitFlags(~_value);
    }

private:
    FlagType _value;
};
}   // namespace MGL
#endif	// INCGUARD_MGL_BIT_H_1610068077

// vim: et ts=4 sw=4 sts=4
