// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_utility.h
 *  \brief      MGL ファイルユーティリティ
 *  \date       Since: September 29, 2022. 14:53:57 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_FILE_UTILITY_H_1664430837
#define INCGUARD_MGL_FILE_UTILITY_H_1664430837

#include <mgl/file/mgl_file_mounter.h>
#include <mgl/stl/mgl_stl_string.h>

namespace MGL::File
{
//! ファイルユーティリティクラス
class Utility
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    constexpr Utility() noexcept
        : _result()
    {
    }
    
    size_t GetSize(const PathView &path) noexcept;
    Result MakeDirectory(const PathView &path) const noexcept;
    Result Move(const PathView &sourcePath, const PathView &destPath) const noexcept;
    Result Remove(const PathView &path) const noexcept;
    Result Copy(const PathView &sourcePath, const PathView &destPath) const noexcept;
    bool Exists(const PathView &path) noexcept;
    bool IsSystemNativeFile(const PathView &path) noexcept;
    Result Mount(const PathView &mountName, const PathView &path, MountAccessType accessType, DelegateKey delegateKey = Mounter::kDefaultDelegateKey) const noexcept;
    Result Unmount(const PathView &mountName) const noexcept;
    Result Remount(const PathView &mountName, const PathView &path, MountAccessType accessType, DelegateKey delegateKey = Mounter::kDefaultDelegateKey) const noexcept;
    bool IsMounted(const PathView &mountName) const noexcept;
    STL::string GetSystemNativePath(const PathView &path) noexcept;
    Result RemoveDelegate(DelegateKey key) const noexcept;
    void SetDefaultDelegate(DelegateKey key) const noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デリゲートの追加
     *  \tparam     DelegateClass       追加するデリゲートクラス
     *  \param[in]  key                 デリゲートキー
     *  \param[in]  args                デリゲートクラスのコンストラクタに渡す引数
     *  \return     失敗時にはいずれかのエラーが設定される
     */
    /* ------------------------------------------------------------------------- */
    template <class DelegateClass, class... Args>
    constexpr Result AddDelegate(DelegateKey key, Args... args) const noexcept
    {
        return Mounter::GetInstance().AddDelegate(key, STL::make_shared<DelegateClass>(args...));
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      処理結果の取得
     *  \return     最後に実行した処理の処理結果
     */
    /* ------------------------------------------------------------------------- */
    constexpr const Result &GetResult() const noexcept
    {
        return _result;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      最後の処理でエラーが発生しているかの取得
     *  \retval     true    エラーが発生している
     *  \retval     false   エラーは発生していない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool HasError() const noexcept
    {
        return _result.HasError();
    }

private:
    Result _result;
};
}   // namespace MGL::File

#endif	// INCGUARD_MGL_FILE_UTILITY_H_1664430837

// vim: et ts=4 sw=4 sts=4
